/*
Copyright (c) 2014, Intel Corporation

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * sends an endless stream of generic Layer 4 packets out of
 * <interface>.  This is to test whether or not NON-PTP
 * packets are being timestamped.
 *
 * Modified for testing purposes by
 *		David Ertman <davidx.m.ertman@intel.com>
 *
 * Based off of the program "timestamping" by
 *		Patrick Ohly <patrick.ohly@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. * See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <net/if.h>

#include <asm/types.h>

#include "packets.h"

#define ERR_ARGUMENTS 1
#define ERR_SETUP 2
#define ERR_TX_TIMESTAMP 4
#define ERR_RX_TIMESTAMP 8

static void usage(const int error)
{
	printf("send_Gen Version 1.2.3\n\n");
	printf("\n\nUsage:\n");
	printf("send_L4 [-h | --help] [-?] [-p | --port <port number>] <interface>\n\n");
	printf("-p | --port <port number>  Port number to send out generic UDP packets\n");
	printf("\tif used, port number on hwtstamp_test has to be the same\n");
	printf("where interface is a valid ethernet device\n");
	exit(error);
}

static void bail(const char *error)
{
	printf("%s: %s\n", error, strerror(errno));
	exit(ERR_SETUP);
}

/* trap the alarm signal so that pause() will wake up on receipt */
static void handle_alarm(int s)
{
	return;
}

static int install_handler(int signum, void(*handler)(int))
{
	struct sigaction action;
	sigset_t mask;

	/* Unblock the signal */
	sigemptyset(&mask);
	sigaddset(&mask, signum);
	sigprocmask(SIG_UNBLOCK, &mask, NULL);

	/* Install the signal handler */
	action.sa_handler = handler;
	action.sa_flags = 0;
	sigemptyset(&action.sa_mask);
	sigaction(signum, &action, NULL);

	return 0;
}

static void sendpacket(int sock, struct sockaddr *addr, socklen_t addr_len)
{
	int res;

	res = 0;
	while (res == 0) {
		res = sendto(sock, GEN_PACKET, sizeof(GEN_PACKET), 0,
						addr, addr_len);
	}
	if (res < 0)
		fprintf(stderr, "%s: %s\n", "send", strerror(errno));
}

int main(int argc, char **argv)
{
	int ip_multicast_loop = 0;
	char *interface;
	int enabled = 1;
	int sock;
	int genPort = 64321;
	struct ifreq device;
	struct sockaddr_in addr;
	struct ip_mreq imr;
	struct in_addr iaddr;
	struct timeval next;

	while (1) {
		static struct option long_options[] = {
			{"help", no_argument,       0, 'h'},
			{"port", required_argument, 0, 'p'},
			{0, 0, 0, 0}
		};

		int option_index = 0;
		int c;

		c = getopt_long (argc, argv, "hp:",
				 long_options, &option_index);

		/* no more options found */
		if (c == -1)
			break;

		switch (c) {
		case 'p':
			if (optarg[0] == '-')
				usage(1);
			genPort = atoi(optarg);
			if (genPort <= 1024)
				bail("Port number must be greater than 1024");
			break;
		case 'h':
			usage(0);
		case '?':
		case '0':
		default:
			usage(ERR_ARGUMENTS);
		}

	}

	if (optind >= argc)
		usage(ERR_ARGUMENTS);
	interface = argv[optind];

	sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0)
		bail("socket");

	memset(&device, 0, sizeof(device));
	strncpy(device.ifr_name, interface, sizeof(device.ifr_name));
	if (ioctl(sock, SIOCGIFADDR, &device) < 0)
		bail("getting interface IP address");

	/* bind to generic port */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(genPort); /* some generic port number */
	if (bind(sock,
		 (struct sockaddr *)&addr,
		 sizeof(struct sockaddr_in)) < 0)
		bail("bind");

	/* set multicast group for outgoing packets */
	inet_aton("224.0.0.1", &iaddr); /* all hosts on network */
	addr.sin_addr = iaddr;
	imr.imr_multiaddr.s_addr = iaddr.s_addr;
	imr.imr_interface.s_addr =
		((struct sockaddr_in *)&device.ifr_addr)->sin_addr.s_addr;
	if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_IF,
		       &imr.imr_interface.s_addr, sizeof(struct in_addr)) < 0)
		bail("set multicast");

	/* join multicast group, loop our own packet */
	if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       &imr, sizeof(struct ip_mreq)) < 0)
		bail("join multicast group");

	if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP,
		       &ip_multicast_loop, sizeof(enabled)) < 0) {
		bail("loop multicast");
	}

	/* request IP_PKTINFO for debugging purposes */
	if (setsockopt(sock, SOL_IP, IP_PKTINFO,
		       &enabled, sizeof(enabled)) < 0)
		printf("%s: %s\n", "setsockopt IP_PKTINFO", strerror(errno));

	/* install the signal handler */
	install_handler(SIGALRM, handle_alarm);

	/* send packets forever every second */
	gettimeofday(&next, 0);
	next.tv_sec = (next.tv_sec + 1);
	next.tv_usec = 0;
	while (1) {
		struct timeval now;
		struct itimerval timer;
		long delta_us;

		gettimeofday(&now, 0);
		delta_us = (long)(next.tv_sec - now.tv_sec) * 1000000 +
			(long)(next.tv_usec - now.tv_usec);
		if (delta_us > 0) {
			memset(&timer, 0, sizeof(timer));
			timer.it_value.tv_usec = delta_us;
			setitimer(ITIMER_REAL, &timer, NULL);
			pause();
		} else {
			/* write one packet */
			sendpacket(sock,
				   (struct sockaddr *)&addr,
				   sizeof(addr));
			printf("Sent data at %ld.%09lu or %s",
			       now.tv_sec, now.tv_usec,
			       ctime(&now.tv_sec));
			next.tv_sec++;
			continue;
		}
	}

	return 0;
}
